"""SC Controller - Modifiers.

Modifier is Action that just sits between input and actual action, changing
way how resulting action works.
For example, click() modifier executes action only if pad is pressed.
"""
from __future__ import annotations

import inspect
import itertools
import logging
import time
from collections import OrderedDict, deque
from math import atan2, copysign, cos, sin, sqrt
from math import pi as PI
from typing import Self

from scc.actions import (
	Action,
	AxisAction,
	GyroAbsAction,
	HapticEnabledAction,
	MouseAction,
	NoAction,
	RangeOP,
	WholeHapticAction,
	XYAction,
)
from scc.constants import (
	CPAD,
	CUT,
	FE_PAD,
	FE_STICK,
	LEFT,
	MINIMUM,
	RIGHT,
	STICK,
	STICK_PAD_MAX,
	STICK_PAD_MAX_HALF,
	STICK_PAD_MIN,
	STICKTILT,
	TRIGGER_MAX,
	ControllerFlags,
	HapticPos,
	SCButtons,
)
from scc.controller import HapticData
from scc.tools import clamp, nameof
from scc.uinput import Axes, Rels

log = logging.getLogger("Modifiers")
_ = lambda x: x


class Modifier(Action):
	def __init__(self, *params):
		Action.__init__(self, *params)
		params = list(params)
		for p in params:
			if isinstance(p, Action):
				self.action = p
				params.remove(p)
				break
		else:
			self.action = NoAction()
		self._mod_init(*params)

	def get_compatible_modifiers(self):
		return self.action.get_compatible_modifiers()

	def cancel(self, mapper):
		self.action.cancel(mapper)

	def get_child_actions(self):
		return (self.action,)

	def _mod_init(self):
		"""Initializes modifier with rest of parameters, after action parameter
		was taken from it and stored in self.action
		"""
		# not needed by default

	def _mod_to_string(self, params, multiline, pad):
		"""Add action at end of params list and generate a string."""
		if multiline:
			childstr = self.action.to_string(True, pad + 2)
			if len(params) > 0:
				return "%s%s(%s,%s%s)" % (
					" " * pad,
					self.COMMAND,
					", ".join([nameof(s) for s in params]),
					"\n" if "\n" in childstr else " ",
					childstr,
				)
			return "%s%s(%s)" % (" " * pad, self.COMMAND, childstr.strip())
		childstr = self.action.to_string(False, pad)
		if len(params) > 0:
			return "%s%s(%s, %s)" % (" " * pad, self.COMMAND, ", ".join([nameof(s) for s in params]), childstr)

		return "%s%s(%s)" % (" " * pad, self.COMMAND, childstr)

	def strip_defaults(self):
		"""Overrides Action.strip_defaults; Uses defaults from _mod_init instead
		of __init__, but does NOT include last of original parameters - action.
		"""
		argspec = inspect.getfullargspec(self.__class__._mod_init)
		required_count = len(argspec.args) - len(argspec.defaults) - 1
		l = list(self.parameters[0:-1])
		d = list(argspec.defaults)[0 : len(l)]
		while len(d) and len(l) > required_count and d[-1] == l[-1]:
			d, l = d[:-1], l[:-1]
		return l

	def strip(self):
		return self.action.strip()

	def compress(self) -> Self:
		if self.action:
			self.action = self.action.compress()
		return self

	def __str__(self):
		return "<Modifier '%s', %s>" % (self.COMMAND, self.action)

	__repr__ = __str__


class NameModifier(Modifier):
	"""Simple modifier that sets name for child action.
	Used internally.
	"""

	COMMAND = "name"

	def _mod_init(self, name):
		self.name = name
		if self.action:
			self.action.name = name

	@staticmethod
	def decode(data, a, *b):
		return a.set_name(data[NameModifier.COMMAND])

	def strip(self):
		rv = self.action.strip()
		rv.name = self.name
		return rv

	@staticmethod
	def unstrip(action):
		# Inversion of strip.
		# For action that has name, returns NameModifier wrapping around that
		# action. For everything else returns original action.
		if not isinstance(action, NameModifier):
			if action and action.name:
				return NameModifier(action.name, action)
		return action

	def compress(self):
		return self.strip()

	def describe(self, context):
		return self.name or self.to_string()

	def to_string(self, multiline=False, pad=0):
		return "%s(%s, %s)" % (self.COMMAND, repr(self.name).strip("u"), self.action.to_string(multiline, pad))


class ClickModifier(Modifier):
	# TODO: Rename to 'clicked'
	COMMAND = "click"

	@staticmethod
	def decode(data, a, *b):
		return ClickModifier(a)

	def describe(self, context):
		if context in (Action.AC_STICK, Action.AC_PAD):
			return _("(if pressed)") + "\n" + self.action.describe(context)
		return _("(if pressed)") + " " + self.action.describe(context)

	def to_string(self, multiline=False, pad=0):
		if multiline:
			childstr = self.action.to_string(True, pad + 2)
			if "\n" in childstr:
				return "%s%s(\n%s\n%s)" % (" " * pad, self.COMMAND, childstr, " " * pad)
		return "%s(%s)" % (self.COMMAND, self.action.to_string())

	def strip(self):
		return self.action.strip()

	def compress(self) -> Self:
		self.action = self.action.compress()
		return self

	# For button press & co it's safe to assume that they are being pressed...
	def button_press(self, mapper):
		return self.action.button_press(mapper)

	def button_release(self, mapper):
		return self.action.button_release(mapper)

	def trigger(self, mapper, position, old_position):
		return self.action.trigger(mapper, position, old_position)

	def axis(self, mapper, position, what):
		if what in (STICK, LEFT) and mapper.is_pressed(SCButtons.LPAD):
			if what == STICK:
				mapper.force_event.add(FE_STICK)
			return self.action.axis(mapper, position, what)
		if what in (STICK, LEFT) and mapper.was_pressed(SCButtons.LPAD):
			# Just released
			return self.action.axis(mapper, 0, what)
		if what == CPAD and mapper.is_pressed(SCButtons.CPAD):
			return self.action.axis(mapper, position, what)
		if what == CPAD and mapper.was_pressed(SCButtons.CPAD):
			# Just released
			return self.action.axis(mapper, 0, what)
		if mapper.is_pressed(SCButtons.RPAD):
			# what == RIGHT, last option
			return self.action.axis(mapper, position, what)
		if mapper.was_pressed(SCButtons.RPAD):
			# what == RIGHT, last option, Just released
			return self.action.axis(mapper, 0, what)

	def pad(self, mapper, position, what):
		if what == LEFT and mapper.is_pressed(SCButtons.LPAD):
			if what == STICK:
				mapper.force_event.add(FE_STICK)
			return self.action.pad(mapper, position, what)
		if what == LEFT and mapper.was_pressed(SCButtons.LPAD):
			# Just released
			return self.action.pad(mapper, 0, what)
		if what == CPAD and mapper.is_pressed(SCButtons.CPAD):
			return self.action.pad(mapper, position, what)
		if what == CPAD and mapper.was_pressed(SCButtons.CPAD):
			# Just released
			return self.action.pad(mapper, 0, what)
		if mapper.is_pressed(SCButtons.RPAD):
			# what == RIGHT, there are only two options
			return self.action.pad(mapper, position, what)
		if mapper.was_pressed(SCButtons.RPAD):
			# what == RIGHT, there are only two options, Just released
			return self.action.pad(mapper, 0, what)

	def whole(self, mapper, x, y, what):
		if what in (STICK, LEFT) and mapper.is_pressed(SCButtons.LPAD):
			if what == STICK:
				mapper.force_event.add(FE_STICK)
			return self.action.whole(mapper, x, y, what)
		if what in (STICK, LEFT) and (mapper.was_pressed(SCButtons.LPAD) or mapper.was_pressed(STICKTILT)):
			# Just released
			return self.action.whole(mapper, 0, 0, what)
		if what == RIGHT and mapper.is_pressed(SCButtons.RPAD):
			return self.action.whole(mapper, x, y, what)
		if what == RIGHT and mapper.was_pressed(SCButtons.RPAD):
			# Just released
			return self.action.whole(mapper, 0, 0, what)
		if what == CPAD and mapper.is_pressed(SCButtons.CPAD):
			return self.action.whole(mapper, x, y, what)
		if what == CPAD and mapper.was_pressed(SCButtons.CPAD):
			# Just released
			return self.action.whole(mapper, 0, 0, what)
		# Nothing is pressed, but finger moves over pad
		self.action.whole_blocked(mapper, x, y, what)


class TouchedModifier(Modifier):
	COMMAND = "touched"

	def describe(self, context):
		if context in (Action.AC_STICK, Action.AC_PAD):
			return _("(when %s)" % (self.COMMAND,)) + "\n" + self.action.describe(context)
		return _("(when %s)" % (self.COMMAND,)) + " " + self.action.describe(context)

	def strip(self):
		return self.action.strip()

	def compress(self) -> Self:
		self.action = self.action.compress()
		return self

	def _release(self, mapper):
		return self.action.button_release(mapper)

	def whole(self, mapper, x, y, what):
		if mapper.is_touched(what) and not mapper.was_touched(what):
			self.action.button_press(mapper)
			mapper.schedule(0.02, self._release)


class UntouchedModifier(TouchedModifier):
	COMMAND = "untouched"

	def whole(self, mapper, x, y, what):
		if not mapper.is_touched(what) and mapper.was_touched(what):
			self.action.button_press(mapper)
			mapper.schedule(0.02, self._release)


class PressedModifier(Modifier):
	COMMAND = "pressed"

	def describe(self, context):
		if context in (Action.AC_STICK, Action.AC_PAD):
			return _("(when pressed)") + "\n" + self.action.describe(context)
		return _("(when pressed)") + " " + self.action.describe(context)

	def strip(self):
		return self.action.strip()

	def compress(self) -> Self:
		self.action = self.action.compress()
		return self

	def button_press(self, mapper):
		self.action.button_press(mapper)
		mapper.schedule(0.02, self._release)

	def _release(self, mapper):
		return self.action.button_release(mapper)

	def button_release(self, mapper):
		pass


class ReleasedModifier(PressedModifier):
	COMMAND = "released"

	def describe(self, context):
		if context in (Action.AC_STICK, Action.AC_PAD):
			return _("(when released)") + "\n" + self.action.describe(context)
		return _("(when released)") + " " + self.action.describe(context)

	def button_press(self, mapper):
		pass

	def button_release(self, mapper):
		self.action.button_press(mapper)
		mapper.schedule(0.02, self._release)


class BallModifier(Modifier, WholeHapticAction):
	"""Emulates ball-like movement with inertia and friction.

	Reacts only to "whole" or "axis" inputs and sends generated movements as
	"add" input to child action.
	Target action has to have add(x, y) method defined.

	"""

	COMMAND = "ball"
	PROFILE_KEY_PRIORITY = -6
	HAPTIC_FACTOR = 60.0  # Just magic number

	DEFAULT_FRICTION = 10.0
	DEFAULT_MEAN_LEN = 10
	MIN_LIFT_VELOCITY = 0.2  # If finger is lifter after movement slower than
	# this, roll doesn't happens

	def __init__(self, *params):
		Modifier.__init__(self, *params)
		WholeHapticAction.__init__(self)

	def _mod_init(
		self, friction=DEFAULT_FRICTION, mass=80.0, mean_len=DEFAULT_MEAN_LEN, r=0.02, ampli=65536, degree=40.0,
	):
		self.speed = (1.0, 1.0)
		self.friction = friction
		self._xvel = 0.0
		self._yvel = 0.0
		self._ampli = ampli
		self._degree = degree
		self._radscale = (degree * PI / 180) / ampli
		self._mass = mass
		self._roll_task = None
		self._r = r
		self._I = (2 * self._mass * self._r**2) / 5.0
		self._a = self._r * self.friction / self._I
		self._xvel_dq = deque(maxlen=mean_len)
		self._yvel_dq = deque(maxlen=mean_len)
		self._lastTime = time.time()
		self._old_pos = None

	def set_speed(self, x, y, *a):
		self.speed = (x, y)
		if self.action and (isinstance(self.action, CircularModifier) and hasattr(self.action, "set_speed")):
			self.action.set_speed(x, y, *a)

		# Reset calculated x and y velocities
		self._xvel = 0.0
		self._yvel = 0.0

	def get_speed(self):
		return self.speed

	def get_compatible_modifiers(self):
		return (
			Action.MOD_SENSITIVITY
			| Action.MOD_FEEDBACK
			| Action.MOD_SMOOTH
			| Action.MOD_DEADZONE
			| Modifier.get_compatible_modifiers(self)
		)

	def _stop(self):
		"""Stops rolling of the 'ball'"""
		self._xvel_dq.clear()
		self._yvel_dq.clear()
		if self._roll_task:
			self._roll_task.cancel()
			self._roll_task = None

	def _add(self, dx, dy):
		# Compute instant velocity
		try:
			self._xvel = sum(self._xvel_dq) / len(self._xvel_dq)
			self._yvel = sum(self._yvel_dq) / len(self._yvel_dq)
		except ZeroDivisionError:
			self._xvel = 0.0
			self._yvel = 0.0

		self._xvel_dq.append(dx * self._radscale)
		self._yvel_dq.append(dy * self._radscale)

	def _roll(self, mapper):
		# Compute time step
		t = time.time()
		dt, self._lastTime = t - self._lastTime, t

		# Free movement update velocity and compute movement
		self._xvel_dq.clear()
		self._yvel_dq.clear()

		_hyp = sqrt((self._xvel**2) + (self._yvel**2))
		if _hyp != 0.0:
			_ax = self._a * (abs(self._xvel) / _hyp)
			_ay = self._a * (abs(self._yvel) / _hyp)
		else:
			_ax = self._a
			_ay = self._a

		# Cap friction desceleration
		_dvx = min(abs(self._xvel), _ax * dt)
		_dvy = min(abs(self._yvel), _ay * dt)

		# compute new velocity
		_xvel = self._xvel - copysign(_dvx, self._xvel)
		_yvel = self._yvel - copysign(_dvy, self._yvel)

		# compute displacement
		dx = (((_xvel + self._xvel) / 2) * dt) / self._radscale
		dy = (((_yvel + self._yvel) / 2) * dt) / self._radscale

		self._xvel = _xvel
		self._yvel = _yvel

		self.action.add(mapper, dx * self.speed[0], dy * self.speed[1])
		if dx or dy:
			if self.haptic:
				WholeHapticAction.add(self, mapper, dx, dy)
			self._roll_task = mapper.schedule(0.02, self._roll)

	@staticmethod
	def decode(data, a, *b):
		if data[BallModifier.COMMAND] is True:
			# backwards compatibility
			return BallModifier(a)
		args = list(data[BallModifier.COMMAND])
		args.append(a)
		return BallModifier(*args)

	def describe(self, context):
		if self.name:
			return self.name
		# Special cases just to make GUI look pretty
		if isinstance(self.action, MouseAction):
			return _("Trackball")
		if isinstance(self.action, XYAction):
			if isinstance(self.action.x, AxisAction) and isinstance(self.action.y, AxisAction):
				x, y = self.action.x.parameters[0], self.action.y.parameters[0]
				if x == Axes.ABS_X and y == Axes.ABS_Y:
					return _("Mouse-like LStick")
				return _("Mouse-like RStick")
			if isinstance(self.action.x, MouseAction) and isinstance(self.action.y, MouseAction):
				x, y = self.action.x.parameters[0], self.action.y.parameters[0]
				if x in (Rels.REL_HWHEEL, Rels.REL_WHEEL) and y in (Rels.REL_HWHEEL, Rels.REL_WHEEL):
					return _("Mouse Wheel")

		return _("Ball(%s)") % (self.action.describe(context))

	def to_string(self, multiline=False, pad=0):
		return self._mod_to_string(self.strip_defaults(), multiline, pad)

	def cancel(self, mapper):
		Modifier.cancel(self, mapper)
		self._stop()

	def pad(self, mapper, position, what):
		self.whole(mapper, position, 0, what)

	def change(self, mapper, dx, dy, what):
		if what in (None, STICK) or (mapper.controller_flags() & ControllerFlags.HAS_RSTICK and what == RIGHT):
			return self.action.change(mapper, dx, dy, what)
		if mapper.is_touched(what):
			if mapper.was_touched(what):
				t = time.time()
				dt = t - self._lastTime
				if dt < 0.0075:
					return None
				self._lastTime = t
				self._add(dx / dt, dy / dt)
				self.action.add(mapper, dx, dy)
			else:
				self._stop()
		elif mapper.was_touched(what):
			velocity = sqrt(self._xvel * self._xvel + self._yvel * self._yvel)
			if velocity > BallModifier.MIN_LIFT_VELOCITY:
				self._roll(mapper)

	def whole(self, mapper, x, y, what):
		if (
			mapper.controller_flags() & ControllerFlags.HAS_RSTICK
			and not mapper.controller_flags() & ControllerFlags.IS_DECK
			and what == RIGHT
		):
			return self.action.whole(mapper, x, y, what)
		if mapper.is_touched(what):
			if mapper.is_touched(what) and not mapper.was_touched(what):
				mapper.mouse.clearRemainders()

			if self._old_pos and mapper.was_touched(what):
				t = time.time()
				dt = t - self._lastTime
				if dt < 0.0075:
					return None
				self._lastTime = t
				dx, dy = x - self._old_pos[0], self._old_pos[1] - y
				self._add(dx / dt, dy / dt)
				self.action.add(mapper, dx * self.speed[0], dy * self.speed[1])
			else:
				self._stop()
			self._old_pos = x, y
		elif mapper.was_touched(what):
			self._old_pos = None
			velocity = sqrt(self._xvel * self._xvel + self._yvel * self._yvel)
			if velocity > BallModifier.MIN_LIFT_VELOCITY:
				self._roll(mapper)
		elif what == STICK:
			return self.action.whole(mapper, x, y, what)

	def set_haptic(self, hd):
		if self.action and hasattr(self.action, "set_haptic"):
			self.action.set_haptic(hd)
		else:
			WholeHapticAction.set_haptic(self, hd)

	def get_haptic(self):
		if self.action and hasattr(self.action, "get_haptic"):
			return self.action.get_haptic()
		return WholeHapticAction.get_haptic(self)

	def compress(self) -> CircularModifier | Self:
		# ball(circular(...) has to be turned around
		if isinstance(self.action, CircularModifier):
			cm = self.action
			self.action = cm.action
			cm.action = self
			return cm
		return self


class DeadzoneModifier(Modifier):
	COMMAND = "deadzone"
	JUMP_HARDCODED_LIMIT = 5

	def _mod_init(self, *params):
		if len(params) < 1:
			raise TypeError("Not enough parameters")
		if type(params[0]) is str:
			self.mode = params[0]
			if hasattr(self, "mode_" + self.mode):
				self._convert = getattr(self, "mode_" + self.mode)
			else:
				raise ValueError("Invalid deadzone mode")
			params = params[1:]
			if len(params) < 1:
				raise TypeError("Not enough parameters")
		else:
			# 'cut' mode is default
			self.mode = CUT
			self._convert = self.mode_CUT

		self.lower = int(params[0])
		self.upper = int(params[1]) if len(params) == 2 else STICK_PAD_MAX

	def mode_CUT(self, x, y, range):
		"""If input value is out of deadzone range, output value is zero
		"""
		if y == 0:
			# Small optimalization for 1D input, for example trigger
			return (0 if abs(x) < self.lower or abs(x) > self.upper else x), 0
		distance = sqrt(x * x + y * y)
		if distance < self.lower or distance > self.upper:
			return 0, 0
		return x, y

	def mode_ROUND(self, x, y, range):
		"""If input value bellow deadzone range, output value is zero
		If input value is above deadzone range,
		output value is 1 (or maximum allowed)
		"""
		if y == 0:
			# Small optimalization for 1D input, for example trigger
			if abs(x) > self.upper:
				return copysign(range, x), 0
			return (0 if abs(x) < self.lower else x), 0
		distance = sqrt(x * x + y * y)
		if distance < self.lower:
			return 0, 0
		if distance > self.upper:
			angle = atan2(x, y)
			return range * sin(angle), range * cos(angle)
		return x, y

	def mode_LINEAR(self, x, y, range):
		"""Input value is scaled, so entire output range is covered by
		reduced input range of deadzone.
		"""
		if y == 0:
			# Small optimalization for 1D input, for example trigger
			return copysign(clamp(0, ((x - self.lower) / (self.upper - self.lower)) * range, range), x), 0
		distance = clamp(self.lower, sqrt(x * x + y * y), self.upper)
		distance = (distance - self.lower) / (self.upper - self.lower) * range

		angle = atan2(x, y)
		return distance * sin(angle), distance * cos(angle)

	def mode_MINIMUM(self, x, y, range):
		"""https://github.com/kozec/sc-controller/issues/356
		Inversion of LINEAR; input value is scaled so entire input range is
		mapped to range of deadzone.
		"""
		if y == 0:
			# Small optimalization for 1D input, for example trigger
			if abs(x) < DeadzoneModifier.JUMP_HARDCODED_LIMIT:
				return 0, 0
			return (copysign((float(abs(x)) / range * (self.upper - self.lower)) + self.lower, x), 0)
		distance = sqrt(x * x + y * y)
		if distance < DeadzoneModifier.JUMP_HARDCODED_LIMIT:
			return 0, 0
		distance = (distance / range * (self.upper - self.lower)) + self.lower

		angle = atan2(x, y)
		return distance * sin(angle), distance * cos(angle)

	def _convert_trigger_stick_range(self, position, trigger_range):
		result = clamp(0, (position / trigger_range) * STICK_PAD_MAX, STICK_PAD_MAX)

		return result

	def _convert_stick_trigger_range(self, position):
		result = clamp(0, (position / STICK_PAD_MAX) * TRIGGER_MAX, TRIGGER_MAX)

		return result

	@staticmethod
	def decode(data, a, *b):
		return DeadzoneModifier(
			data["deadzone"]["mode"] if "mode" in data["deadzone"] else CUT,
			data["deadzone"]["lower"] if "lower" in data["deadzone"] else STICK_PAD_MIN,
			data["deadzone"]["upper"] if "upper" in data["deadzone"] else STICK_PAD_MAX,
			a,
		)

	def compress(self) -> BallModifier | GyroAbsAction | Self:
		self.action = self.action.compress()
		if isinstance(self.action, BallModifier) and self.mode == MINIMUM:
			# Special case where BallModifier has to be applied before
			# deadzone is computed
			ballmod = self.action
			self.action, ballmod.action = ballmod.action, self
			return ballmod
		if isinstance(self.action, GyroAbsAction):
			# Another special case, GyroAbs has to handle deadzone
			# only after math is finished
			self.action._deadzone_fn = self._convert
			return self.action
		return self

	def strip(self):
		return self.action.strip()

	def __str__(self):
		return "<Modifier '%s', %s>" % (self.COMMAND, self.action)

	__repr__ = __str__

	def describe(self, context):
		dsc = self.action.describe(context)
		if "\n" in dsc:
			return "%s\n(with deadzone)" % (dsc,)
		return "%s (with deadzone)" % (dsc,)

	def to_string(self, multiline=False, pad=0):
		params = []
		if self.mode != CUT:
			params.append(self.mode)
		params.append(str(self.lower))
		if self.upper != STICK_PAD_MAX:
			params.append(str(self.upper))
		params.append(self.action.to_string(multiline))

		return "deadzone(%s)" % (", ".join(params),)

	def trigger(self, mapper, position, old_position):
		# Need to convert trigger value to stick range for deadzone modifier
		# calcs to work as intended
		position = self._convert_trigger_stick_range(position, TRIGGER_MAX)

		# Perform dead zone calculations
		position = self._convert(position, 0, STICK_PAD_MAX)

		# Convert calculated stick position value back to applicable
		# position in trigger range
		position = self._convert_stick_trigger_range(position[0])
		# Invoke trigger action with new value
		return self.action.trigger(mapper, position, old_position)

	def axis(self, mapper, position, what):
		position = self._convert(position, 0, STICK_PAD_MAX)
		return self.action.axis(mapper, position, what)

	def pad(self, mapper, position, what):
		position = self._convert(position, 0, STICK_PAD_MAX)
		return self.action.pad(mapper, position, what)

	def whole(self, mapper, x, y, what):
		x, y = self._convert(x, y, STICK_PAD_MAX)
		return self.action.whole(mapper, x, y, what)

	def gyro(self, mapper, pitch, yaw, roll, q1, q2, q3, q4):
		return self.action.gyro(mapper, pitch, yaw, roll, q1, q2, q3, q4)


class ModeModifier(Modifier):
	COMMAND = "mode"
	PROFILE_KEYS = ("modes",)
	MIN_TRIGGER = 2  # When trigger is bellow this position, list of held_triggers is cleared
	MIN_STICK = 2  # When abs(stick) < MIN_STICK, stick is considered released and held_sticks is cleared
	PROFILE_KEY_PRIORITY = 2

	def __init__(self, *stuff):
		# TODO: Better documentation for this. For now, using shell
		# TODO: and range as condition is not documented
		Modifier.__init__(self)
		self.default = None
		self.mods = OrderedDict()
		self.held_buttons = set()
		self.held_sticks = set()
		self.held_triggers = {}
		self.old_action = None
		self.shell_commands = {}
		self.shell_timeout = 0.5
		self.timeout = DoubleclickModifier.DEFAULT_TIMEOUT

		# ShellCommandAction cannot be imported normally, it would create
		# import cycle of hell
		ShellCommandAction = Action.ALL["shell"]
		button = None
		for i in stuff:
			if self.default is not None:
				# Default has to be last parameter
				raise ValueError("Invalid parameters for 'mode'")
			if isinstance(i, ShellCommandAction) and button is None:
				# 'shell' can be used instead of button
				button = i
			elif isinstance(i, Action) and button is None:
				self.default = i
			elif isinstance(i, Action):
				self.mods[button] = i
				button = None
			elif isinstance(i, RangeOP) or i in SCButtons.__members__.values():
				button = i
			else:
				raise ValueError("Invalid parameter for 'mode': %s" % (i,))
		self.make_checks()
		if self.default is None:
			if isinstance(button, ShellCommandAction):
				self.default = button
			else:
				self.default = NoAction()

	def make_checks(self):
		self.checks = []
		self.shell_commands = {}
		ShellCommandAction = Action.ALL["shell"]
		for c, action in self.mods.items():
			if isinstance(c, RangeOP):
				self.checks.append((c, action))
			elif isinstance(c, ShellCommandAction):
				self.shell_commands[c.command] = c
				self.checks.append((self.make_shell_check(c), action))
			else:
				self.checks.append((self.make_button_check(c), action))

	def get_child_actions(self):
		rv = list(self.mods.values()) + list(self.shell_commands.values())
		if self.default is not None:
			rv += [self.default]
		return rv

	@staticmethod
	def decode(data, a, parser, *b):
		args = []
		for button in data[ModeModifier.PROFILE_KEYS[0]]:
			if hasattr(SCButtons, button):
				args += [getattr(SCButtons, button), parser.from_json_data(data[ModeModifier.PROFILE_KEYS[0]][button])]
		if a:
			args += [a]
		mm = ModeModifier(*args)
		if "name" in data:
			mm.name = data["name"]
		return mm

	def get_compatible_modifiers(self):
		rv = 0
		for action in self.mods.values():
			rv |= action.get_compatible_modifiers()
		if self.default:
			rv |= self.default.get_compatible_modifiers()
		return rv

	def strip(self):
		# Returns default action or action assigned to first modifier
		if self.default:
			return self.default.strip()
		if len(self.mods):
			return next(itertools.islice(self.mods.values(), 0, 1)).strip()
		# Empty ModeModifier
		return NoAction()

	def compress(self) -> Self:
		if self.default:
			self.default = self.default.compress()
		for check in self.mods:
			self.mods[check] = self.mods[check].compress()
		self.make_checks()
		return self

	def __str__(self):
		rv = []
		for check in self.mods:
			rv += [nameof(check), self.mods[check]]
		if self.default is not None:
			rv += [self.default]
		return "<Modifier '%s', %s>" % (self.COMMAND, rv)

	__repr__ = __str__

	def describe(self, context):
		if self.name:
			return self.name
		l = []
		if self.default:
			l.append(self.default)
		for check in self.mods:
			l.append(self.mods[check])
		return "\n".join([x.describe(context) for x in l])

	def to_string(self, multiline: bool = False, pad=0):
		if multiline:
			rv = [(" " * pad) + "mode("]
			for check in self.mods:
				a_str = NameModifier.unstrip(self.mods[check]).to_string(True).split("\n")
				a_str[0] = (
					(" " * pad) + "  " + (nameof(check) + ",").ljust(11) + a_str[0]
				)  # Key has to be one of SCButtons
				for i in range(1, len(a_str)):
					a_str[i] = (" " * pad) + "  " + a_str[i]
				a_str[-1] = a_str[-1] + ","
				rv += a_str
			if self.default is not None:
				a_str = [(" " * pad) + "  " + x for x in NameModifier.unstrip(self.default).to_string(True).split("\n")]
				rv += a_str
			if rv[-1][-1] == ",":
				rv[-1] = rv[-1][0:-1]
			rv += [(" " * pad) + ")"]
			return "\n".join(rv)
		rv = []
		for check in self.mods:
			rv += [nameof(check), NameModifier.unstrip(self.mods[check]).to_string(False)]
		if self.default is not None:
			rv += [NameModifier.unstrip(self.default).to_string(False)]
		return "mode(" + ", ".join(rv) + ")"

	def cancel(self, mapper):
		for action in self.mods.values():
			action.cancel(mapper)
		self.default.cancel(mapper)

	def select(self, mapper):
		"""Selects action by pressed button.
		"""
		for check, action in self.checks:
			if check(mapper):
				return action
		return self.default

	def select_w_check(self, mapper):
		"""As select, but returns matched check as well.
		"""
		for check, action in self.checks:
			if check(mapper):
				return check, action
		return lambda *a: True, self.default

	@staticmethod
	def make_button_check(button):
		def cb(mapper):
			return mapper.is_pressed(button)

		cb.name = button.name  # So nameof() still works on keys in self.mods
		return cb

	@staticmethod
	def make_shell_check(c):
		def cb(mapper):
			try:
				return c.__proc.poll() == 0
			except:
				return False

		c.name = cb.name = c.to_string()  # So nameof() still works on keys in self.mods
		c.__proc = None
		return cb

	def button_press(self, mapper):
		if len(self.shell_commands) > 0:
			# https://github.com/kozec/sc-controller/issues/427
			# If 'shell' is used as any condition, all shell commands
			# are executed and ModeShift waits up to 500ms for them
			# to terminate. Then, if command returned zero exit code
			# it's considered as 'true' condition.
			for c in self.shell_commands.values():
				c.__proc = c.button_press(mapper)
			self.shell_timeout = 0.5
			mapper.schedule(0, self.check_shell_commands)
			return None

		sel = self.select(mapper)
		self.held_buttons.add(sel)
		return sel.button_press(mapper)

	def check_shell_commands(self, mapper):
		for c in self.shell_commands.values():
			if c.__proc and c.__proc.poll() == 0:
				sel = self.select(mapper)
				self.kill_shell_commands()
				self.held_buttons.add(sel)
				return sel.button_press(mapper)

		self.shell_timeout -= 0.05
		if self.shell_timeout > 0:
			mapper.schedule(0.05, self.check_shell_commands)
		else:
			# time is up, kill all processes and execute what's left
			self.kill_shell_commands()
			sel = self.select(mapper)
			self.held_buttons.add(sel)
			return sel.button_press(mapper)

	def kill_shell_commands(self):
		for c in self.shell_commands.values():
			try:
				if c.__proc:
					c.__proc.kill()
			except:
				pass
			c.__proc = None

	def button_release(self, mapper):
		# Releases all held buttons, not just button that matches
		# currently pressed modifier
		for b in self.held_buttons:
			b.button_release(mapper)

	def trigger(self, mapper, position, old_position):
		if position < ModeModifier.MIN_TRIGGER:
			for b in self.held_triggers:
				b.trigger(mapper, 0, self.held_triggers[b])
			self.held_triggers = {}
			return False
		sel = self.select(mapper)
		self.held_triggers[sel] = position
		return sel.trigger(mapper, position, old_position)

	def axis(self, mapper, position, what):
		return self.select(mapper).axis(mapper, position, what)

	def gyro(self, mapper, pitch, yaw, roll, *q):
		sel = self.select(mapper)
		if sel is not self.old_action:
			if self.old_action:
				self.old_action.gyro(mapper, 0, 0, 0, *q)
			self.old_action = sel
		return sel.gyro(mapper, pitch, yaw, roll, *q)

	def pad(self, mapper, position, what):
		return self.select(mapper).pad(mapper, position, what)

	def whole(self, mapper, x, y, what):
		if what == STICK:
			if abs(x) < ModeModifier.MIN_STICK and abs(y) < ModeModifier.MIN_STICK:
				for check, action in self.held_sticks:
					action.whole(mapper, 0, 0, what)
				self.held_sticks.clear()
			else:
				ac, active = self.select_w_check(mapper)
				self.held_sticks.add((ac, active))
				for check, action in list(self.held_sticks):
					if check == ac or check(mapper):
						action.whole(mapper, x, y, what)
					else:
						action.whole(mapper, 0, 0, what)
						self.held_sticks.remove((check, action))
			mapper.force_event.add(FE_STICK)
		else:
			sel = self.select(mapper)
			if sel is not self.old_action:
				mapper.set_button(what, False)
				if self.old_action:
					self.old_action.whole(mapper, 0, 0, what)
				self.old_action = sel
				rv = sel.whole(mapper, x, y, what)
				mapper.set_button(what, True)
				return rv
			return sel.whole(mapper, x, y, what)


class DoubleclickModifier(Modifier, HapticEnabledAction):
	COMMAND = "doubleclick"
	DEFAULT_TIMEOUT = 0.2
	TIMEOUT_KEY = "time"
	PROFILE_KEY_PRIORITY = 3

	def __init__(self, doubleclickaction, normalaction=None, time=None):
		Modifier.__init__(self)
		HapticEnabledAction.__init__(self)
		self.action = doubleclickaction
		self.normalaction = normalaction or NoAction()
		self.holdaction = NoAction()
		self.actions = (self.action, self.normalaction, self.holdaction)
		self.timeout = time or DoubleclickModifier.DEFAULT_TIMEOUT
		self.waiting_task = None
		self.pressed = False
		self.active = None

	def get_child_actions(self):
		return self.action, self.normalaction, self.holdaction

	@staticmethod
	def decode(data, a, parser, *b):
		args = [parser.from_json_data(data[DoubleclickModifier.COMMAND]), a]
		a = DoubleclickModifier(*args)
		if DoubleclickModifier.TIMEOUT_KEY in data:
			a.timeout = data[DoubleclickModifier.TIMEOUT_KEY]
		return a

	def strip(self):
		if self.holdaction:
			return self.holdaction.strip()
		return self.action.strip()

	def compress(self) -> Self:
		self.action = self.action.compress()
		self.holdaction = self.holdaction.compress()
		self.normalaction = self.normalaction.compress()

		for a in (self.holdaction, self.normalaction):
			if isinstance(a, HoldModifier):
				self.holdaction = a.holdaction or self.holdaction
				self.normalaction = a.normalaction or self.normalaction

		if isinstance(self.action, HoldModifier):
			self.holdaction = self.action.holdaction
			self.action = self.action.normalaction
		return self

	def __str__(self):
		l = [self.action]
		if self.normalaction:
			l += [self.normalaction]
		return "<Modifier %s dbl='%s' hold='%s' normal='%s'>" % (
			self.COMMAND,
			self.action,
			self.holdaction,
			self.normalaction,
		)

	__repr__ = __str__

	def describe(self, context):
		l = []
		if self.action:
			l += [self.action]
		if self.holdaction:
			l += [self.holdaction]
		if self.normalaction:
			l += [self.normalaction]
		return "\n".join([x.describe(context) for x in l])

	def to_string(self, multiline=False, pad=0):
		timeout = ""
		if self.timeout != DoubleclickModifier.DEFAULT_TIMEOUT:
			timeout = ", %s" % (self.timeout)
		if self.action and self.normalaction and self.holdaction:
			return "doubleclick(%s, hold(%s, %s)%s)" % (
				NameModifier.unstrip(self.action).to_string(multiline, pad),
				NameModifier.unstrip(self.holdaction).to_string(multiline, pad),
				NameModifier.unstrip(self.normalaction).to_string(multiline, pad),
				timeout,
			)
		if self.action and self.normalaction and not self.holdaction:
			return "doubleclick(%s, %s%s)" % (
				NameModifier.unstrip(self.action).to_string(multiline, pad),
				NameModifier.unstrip(self.normalaction).to_string(multiline, pad),
				timeout,
			)
		if not self.action and self.normalaction and self.holdaction:
			return "hold(%s, %s%s)" % (
				NameModifier.unstrip(self.holdaction).to_string(multiline, pad),
				NameModifier.unstrip(self.normalaction).to_string(multiline, pad),
				timeout,
			)
		if not self.action and not self.normalaction and self.holdaction:
			return "hold(None, %s%s)" % (NameModifier.unstrip(self.holdaction).to_string(multiline, pad), timeout)
		if self.action and not self.normalaction and not self.holdaction:
			return "doubleclick(None, %s%s)" % (NameModifier.unstrip(self.action).to_string(multiline, pad), timeout)
		return NameModifier.unstrip(self.action or self.normalaction or self.holdaction).to_string(multiline, pad)

	def button_press(self, mapper):
		self.pressed = True
		if self.waiting_task:
			# Double-click happened
			mapper.cancel_task(self.waiting_task)
			self.waiting_task = None
			self.active = self.action
			self.active.button_press(mapper)
		else:
			# First click, start the timer
			self.waiting_task = mapper.schedule(self.timeout, self.on_timeout)

	def button_release(self, mapper):
		self.pressed = False
		if self.waiting_task and self.active is None and not self.action:
			# In HoldModifier, button released before timeout
			mapper.cancel_task(self.waiting_task)
			self.waiting_task = None
			if self.normalaction:
				self.normalaction.button_press(mapper)
				mapper.schedule(0.02, self.normalaction.button_release)
		elif self.active:
			# Released held button
			self.active.button_release(mapper)
			self.active = None

	def on_timeout(self, mapper, *a):
		if self.waiting_task:
			self.waiting_task = None
			if self.pressed:
				# Timeouted while button is still pressed
				self.active = self.holdaction or self.normalaction
				if self.haptic:
					mapper.send_feedback(self.haptic)
				self.active.button_press(mapper)
			elif self.normalaction:
				# User did short click and nothing else
				self.normalaction.button_press(mapper)
				mapper.schedule(0.02, self.normalaction.button_release)


class HoldModifier(DoubleclickModifier):
	# Hold modifier is implemented as part of DoubleclickModifier, because
	# situation when both are assigned to same button needs to be treated
	# specially.
	COMMAND = "hold"
	PROFILE_KEY_PRIORITY = 4

	def __init__(self, holdaction, normalaction=None, time=None):
		DoubleclickModifier.__init__(self, NoAction(), normalaction, time)
		self.holdaction = holdaction

	@staticmethod
	def decode(data, a, parser, *b):
		if isinstance(a, DoubleclickModifier):
			a.holdaction = parser.from_json_data(data[HoldModifier.COMMAND])
		else:
			args = [parser.from_json_data(data[HoldModifier.COMMAND]), a]
			a = HoldModifier(*args)
		if DoubleclickModifier.TIMEOUT_KEY in data:
			a.timeout = data[DoubleclickModifier.TIMEOUT_KEY]
		if isinstance(a.normalaction, FeedbackModifier):
			# Ugly hack until profile file is redone
			mod = a.normalaction
			a.normalaction = mod.action
			if hasattr(a.normalaction, "set_haptic"):
				a.normalaction.set_haptic(None)
			mod.action = a
			mod.action.set_haptic(mod.haptic)
			a = mod
		return a

	def compress(self) -> Self:
		self.action = self.action.compress()
		self.holdaction = self.holdaction.compress()
		self.normalaction = self.normalaction.compress()

		for a in (self.action, self.normalaction):
			if isinstance(a, DoubleclickModifier):
				self.action = a.action or self.action
				self.normalaction = a.normalaction or self.normalaction

		if isinstance(self.holdaction, DoubleclickModifier):
			self.action = self.holdaction.action
			self.holdaction = self.holdaction.normalaction
		return self


class SensitivityModifier(Modifier):
	"""Sets action sensitivity, if action supports it.
	Action that supports such setting has set_speed(x, y, z)
	and get_speed() methods defined.

	Does nothing otherwise.
	"""

	COMMAND = "sens"
	PROFILE_KEYS = ("sensitivity",)
	PROFILE_KEY_PRIORITY = -5

	def _mod_init(self, *speeds) -> None:
		self.speeds = []
		for s in speeds:
			if type(s) in (int, float):
				self.speeds.append(float(s))
		while len(self.speeds) < 3:
			self.speeds.append(1.0)
		if self.action:
			a = self.action
			while a:
				if hasattr(a, "set_speed"):
					a.set_speed(*self.speeds)
					break
				if hasattr(a, "action"):
					a = a.action
				else:
					break

	@staticmethod
	def decode(data, a, *b):
		if a:
			args = list(data["sensitivity"])
			args.append(a)
			return SensitivityModifier(*args)
		# Adding sensitivity to NoAction makes no sense
		return a

	def strip(self):
		return self.action.strip()

	def compress(self):
		return self.action.compress()

	def describe(self, context):
		if self.name:
			return self.name
		return self.action.describe(context)

	def to_string(self, multiline: bool = False, pad=0) -> str:
		speeds = [] + self.speeds
		while len(speeds) > 1 and speeds[-1] == 1.0:
			speeds = speeds[0:-1]
		return self._mod_to_string(speeds, multiline, pad)

	def __str__(self):
		return "<Sensitivity=%s, %s>" % (self.speeds, self.action)


class FeedbackModifier(Modifier):
	"""Enables feedback for action, action supports it.
	Action that supports feedback has to have set_haptic(hapticdata)
	method defined.

	Does nothing otherwise.
	"""

	COMMAND = "feedback"
	PROFILE_KEY_PRIORITY = -4

	def _mod_init(self, position, amplitude=512, frequency=4, period=1024, count=1):
		self.haptic = HapticData(position, amplitude, frequency, period, count)
		if self.action:
			a = self.action
			while a:
				if hasattr(a, "set_haptic"):
					a.set_haptic(self.haptic)
					break
				if hasattr(a, "action"):
					a = a.action
				else:
					break

	@staticmethod
	def decode(data, a, *b):
		args = list(data[FeedbackModifier.COMMAND])
		if hasattr(HapticPos, args[0]):
			args[0] = getattr(HapticPos, args[0])
		args.append(a)
		return FeedbackModifier(*args)

	def describe(self, context):
		if self.name:
			return self.name
		return self.action.describe(context)

	def to_string(self, multiline=False, pad=0):
		return self._mod_to_string(self.strip_defaults(), multiline, pad)

	def __str__(self):
		return "<with Feedback %s>" % (self.action,)

	def strip(self):
		return self.action.strip()

	def compress(self):
		return self.action.compress()


class RotateInputModifier(Modifier):
	"""Rotates ball or stick input along axis"""

	COMMAND = "rotate"

	def _mod_init(self, angle):
		self.angle = angle

	@staticmethod
	def decode(data, a, *b):
		return RotateInputModifier(float(data["rotate"]), a)

	def describe(self, context):
		if self.name:
			return self.name
		return self.action.describe(context)

	def to_string(self, multiline=False, pad=0):
		return self._mod_to_string((self.angle,), multiline, pad)

	def strip(self):
		return self.action.strip()

	def compress(self) -> Action | NoAction | Self:
		if hasattr(self.action, "set_rotation"):
			self.action.set_rotation(self.angle * PI / -180.0)
			return self.action
		self.action = self.action.compress()
		return self

	# This doesn't make sense with anything but 'whole' as input.
	def whole(self, mapper, x, y, what):
		angle = self.angle * PI / -180.0
		rx = x * cos(angle) - y * sin(angle)
		ry = x * sin(angle) + y * cos(angle)
		return self.action.whole(mapper, rx, ry, what)


class SmoothModifier(Modifier):
	"""Smooths pad movements
	"""

	COMMAND = "smooth"
	PROFILE_KEY_PRIORITY = 11  # Before sensitivity

	def _mod_init(self, level=8, multiplier=0.75, filter=2.0):
		self.level = level
		self.multiplier = multiplier
		self.filter = filter
		self._deq_x = deque([0.0] * level, maxlen=level)
		self._deq_y = deque([0.0] * level, maxlen=level)
		self._range = list(range(level))
		self._weights = [multiplier**x for x in reversed(self._range)]
		self._w_sum = sum(self._weights)
		self._last_pos = None
		self._moving = False

	def __str__(self):
		return "<Smooth %s>" % (self.action,)

	def describe(self, context):
		if self.name:
			return self.name
		return "%s (smooth)" % (self.action.describe(context),)

	@staticmethod
	def decode(data, a, *b):
		pars = data[SmoothModifier.COMMAND] + [a]
		return SmoothModifier(*pars)

	def _get_pos(self):
		"""Computes average x,y from all accumulated positions"""
		x = sum(self._deq_x[i] * self._weights[i] for i in self._range)
		y = sum(self._deq_y[i] * self._weights[i] for i in self._range)
		return int(x / self._w_sum), int(y / self._w_sum)

	def whole(self, mapper, x, y, what):
		if mapper.controller_flags() & ControllerFlags.HAS_RSTICK and what == RIGHT:
			return self.action.whole(mapper, x, y, what)
		if mapper.is_touched(what):
			if self._last_pos is None:
				# Just pressed - fill deque with current position
				for i in self._range:
					self._deq_x.append(x)
					self._deq_y.append(y)
				x, y = self._get_pos()
				self._last_pos = 0
			else:
				# Pressed for longer time
				self._deq_x.append(x)
				self._deq_y.append(y)
				x, y = self._get_pos()
			if abs(x + y - self._last_pos) > self.filter:
				self.action.whole(mapper, x, y, what)
			self._last_pos = x + y
		elif what == STICK:
			return self.action.whole(mapper, x, y, what)
		else:
			# Pad was just released
			x, y = self._get_pos()
			self.action.whole(mapper, x, y, what)
			self._last_pos = None


class CircularModifier(Modifier, HapticEnabledAction):
	"""Designed to translate rotating finger over pad to mouse wheel movement.
	Can also be used to translate same thing into movement of Axis.
	"""

	COMMAND = "circular"
	PROFILE_KEY_PRIORITY = -6

	def __init__(self, *params):
		# Piece of backwards compatibility
		if len(params) >= 1 and params[0] in Rels.__members__.values():
			params = [MouseAction(params[0])]
		self._haptic_counter = 0
		Modifier.__init__(self, *params)
		HapticEnabledAction.__init__(self)

	def _mod_init(self):
		self.angle = None  # Last known finger position
		self.speed = 1.0

	def set_haptic(self, hd):
		if isinstance(self.action, HapticEnabledAction):
			self.action.set_haptic(hd)
		else:
			HapticEnabledAction.set_haptic(self, hd)

	def get_haptic(self):
		if isinstance(self.action, HapticEnabledAction):
			return self.action.get_haptic()
		return HapticEnabledAction.get_haptic(self)

	@staticmethod
	def decode(data, a, *b):
		return CircularModifier(a)

	def describe(self, context):
		if self.name:
			return self.name
		return _("Circular %s") % (self.action.describe(context))

	def set_speed(self, x, *a):
		self.speed = x

	def get_speed(self):
		return (self.speed,)

	def get_compatible_modifiers(self):
		return Action.MOD_FEEDBACK | Action.MOD_SENSITIVITY | Modifier.get_compatible_modifiers(self)

	def whole(self, mapper, x, y, what):
		distance = sqrt(x * x + y * y)
		if distance < STICK_PAD_MAX_HALF:
			# Finger lifted or too close to middle
			self.angle = None
			if mapper.was_touched(what):
				self.action.change(mapper, 0, 0, what)
		else:
			# Compute current angle
			angle = atan2(x, y)
			# Compute movement
			if self.angle is None:
				# Finger just touched the pad
				self.angle, angle = angle, 0
				self._haptic_counter = 0
			else:
				self.angle, angle = angle, self.angle - angle
				# Ensure we don't wrap from pi to -pi creating a large delta
				if angle > PI:
					# Subtract a full rotation to counter the wrapping
					angle -= 2 * PI
				# And same from -pi to pi
				elif angle < -PI:
					# Add a full rotation to counter the wrapping
					angle += 2 * PI
			# Apply bulgarian constant
			angle *= 10000.0
			# Generate feedback, if enabled
			if self.haptic:
				self._haptic_counter += angle * self.speed / self.haptic.frequency
				if abs(self._haptic_counter) > 0.5:
					if self._haptic_counter > 0.5:
						self._haptic_counter -= 0.5
					else:
						self._haptic_counter += 0.5
					mapper.send_feedback(self.haptic)
			# Apply movement to child action
			# Keep event from activating if no angle change
			if angle != 0.0:
				self.action.change(mapper, -angle * self.speed, 0, what)
				mapper.force_event.add(FE_PAD)


class CircularAbsModifier(Modifier, WholeHapticAction):
	"""Works similary to CircularModifier, but instead of counting with finger
	movements movements, translates exact position on dpad to axis value.
	"""

	COMMAND = "circularabs"
	PROFILE_KEY_PRIORITY = -6

	def __init__(self, *params):
		Modifier.__init__(self, *params)
		WholeHapticAction.__init__(self)

	def _mod_init(self):
		self.angle = None  # Last known finger position
		self.speed = 1.0

	@staticmethod
	def decode(data, a, *b):
		return CircularAbsModifier(a)

	def describe(self, context):
		if self.name:
			return self.name
		return _("Absolute Circular %s") % (self.action.describe(context))

	def set_speed(self, x, *a):
		self.speed = x

	def get_speed(self):
		return (self.speed,)

	def get_compatible_modifiers(self):
		return (
			Action.MOD_FEEDBACK | Action.MOD_SENSITIVITY | Action.MOD_ROTATE | Modifier.get_compatible_modifiers(self)
		)

	def whole(self, mapper, x, y, what):
		distance = sqrt(x * x + y * y)
		if distance < STICK_PAD_MAX_HALF:
			# Finger lifted or too close to middle
			self.angle = None
		else:
			# Compute current angle
			angle = atan2(x, y) + PI / 4
			# Compute movement
			if self.haptic:
				if self.angle is not None:
					diff = self.angle - angle
					# Ensure we don't wrap from pi to -pi creating a large delta
					if angle > PI:
						# Subtract a full rotation to counter the wrapping
						angle -= 2 * PI
					# And same from -pi to pi
					elif angle < -PI:
						# Add a full rotation to counter the wrapping
						angle += 2 * PI
					if abs(diff) < 6:  # Prevents crazy feedback burst when finger cross 360' angle
						WholeHapticAction.change(self, mapper, diff * 10000, 0, what)
				self.angle = angle
			# Apply actual constant
			angle *= STICK_PAD_MAX / PI
			# Set axis on child action
			self.action.axis(mapper, angle * self.speed, 0)
			mapper.force_event.add(FE_PAD)



"""SC Controller - Poller

Uses select to pool for file descriptors. Driver classes can use
daemon.get_poller().register and .unregister to add file descriptors and
register callbacks to be called when data is available in them.

Callback is called as callback(fd, event) where event is one of select.POLL*
"""

import logging
import select

log = logging.getLogger("Poller")


DO_NOTHING = lambda *a: False


class Poller:
	POLLIN = select.POLLIN
	POLLOUT = select.POLLOUT
	POLLPRI = select.POLLPRI

	def __init__(self) -> None:
		self._events = {}
		self._callbacks = {}
		self._pool_in = ()
		self._pool_out = ()
		self._pool_pri = ()

	def register(self, fd, events, callback) -> None:
		if fd < 0:
			raise ValueError("Invalid file descriptor")
		self._events[fd] = events
		self._callbacks[fd] = callback
		self._generate_lists()

	def unregister(self, fd) -> None:
		if fd in self._events:
			del self._events[fd]
		if fd in self._callbacks:
			del self._callbacks[fd]
		self._generate_lists()

	def _generate_lists(self) -> None:
		self._pool_in = [fd for fd, events in self._events.items() if events & Poller.POLLIN]
		self._pool_out = [fd for fd, events in self._events.items() if events & Poller.POLLOUT]
		self._pool_pri = [fd for fd, events in self._events.items() if events & Poller.POLLPRI]

	def poll(self, timeout=0.01) -> None:
		inn, out, pri = select.select(self._pool_in, self._pool_out, self._pool_pri, timeout)

		for fd in inn:
			self._callbacks.get(fd, DO_NOTHING)(fd, Poller.POLLIN)
		for fd in out:
			self._callbacks.get(fd, DO_NOTHING)(fd, Poller.POLLOUT)
		for fd in pri:
			self._callbacks.get(fd, DO_NOTHING)(fd, Poller.POLLPRI)



"""udevmonitor.py - enumerates and monitors devices using (e)udev.

Copyright (C) 2018 by Kozec

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2 as published by
the Free Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from __future__ import annotations

import ctypes
import errno
import os
from ctypes.util import find_library
from typing import NamedTuple, Self, TypeVar


class Eudev:
	"""Wrapped udev system library."""

	LIB_NAME = "udev"

	def __init__(self) -> None:
		self._ctx = None
		try:
			self._lib = ctypes.cdll.LoadLibrary("libudev.so.1")
		except OSError:
			lib_name = find_library(self.LIB_NAME)
			if lib_name is None:
				raise ImportError("No library named udev")
			self._lib = ctypes.CDLL(lib_name)
		Eudev._setup_lib(self._lib)
		self._ctx = self._lib.udev_new()
		if self._ctx is None:
			raise OSError("Failed to initialize udev context")

	@staticmethod
	def _setup_lib(lib: ctypes.CDLL) -> None:
		"""Just so it's away from init and can be folded in IDE."""
		# udev
		lib.udev_new.restype = ctypes.c_void_p
		lib.udev_unref.argtypes = [ctypes.c_void_p]
		# enumeration
		lib.udev_enumerate_new.argtypes = [ctypes.c_void_p]
		lib.udev_enumerate_new.restype = ctypes.c_void_p
		lib.udev_enumerate_unref.argtypes = [ctypes.c_void_p]
		lib.udev_enumerate_scan_devices.argtypes = [ctypes.c_void_p]
		lib.udev_enumerate_scan_devices.restype = ctypes.c_int
		lib.udev_enumerate_get_list_entry.argtypes = [ctypes.c_void_p]
		lib.udev_enumerate_get_list_entry.restype = ctypes.c_void_p
		lib.udev_list_entry_get_next.argtypes = [ctypes.c_void_p]
		lib.udev_list_entry_get_next.restype = ctypes.c_void_p
		lib.udev_list_entry_get_value.argtypes = [ctypes.c_void_p]
		lib.udev_list_entry_get_value.restype = ctypes.c_char_p
		lib.udev_list_entry_get_name.argtypes = [ctypes.c_void_p]
		lib.udev_list_entry_get_name.restype = ctypes.c_char_p
		# monitoring
		lib.udev_monitor_new_from_netlink.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
		lib.udev_monitor_new_from_netlink.restype = ctypes.c_void_p
		lib.udev_monitor_unref.argtypes = [ctypes.c_void_p]
		lib.udev_monitor_enable_receiving.argtypes = [ctypes.c_void_p]
		lib.udev_monitor_enable_receiving.restype = ctypes.c_int
		lib.udev_monitor_set_receive_buffer_size.argtypes = [ctypes.c_void_p, ctypes.c_int]
		lib.udev_monitor_set_receive_buffer_size.restype = ctypes.c_int
		lib.udev_monitor_get_fd.argtypes = [ctypes.c_void_p]
		lib.udev_monitor_get_fd.restype = ctypes.c_int
		lib.udev_monitor_receive_device.argtypes = [ctypes.c_void_p]
		lib.udev_monitor_receive_device.restype = ctypes.c_void_p
		lib.udev_monitor_filter_update.argtypes = [ctypes.c_void_p]
		lib.udev_monitor_filter_update.restype = ctypes.c_int
		lib.udev_monitor_filter_add_match_subsystem_devtype.argtypes = [
			ctypes.c_void_p,
			ctypes.c_char_p,
			ctypes.c_char_p,
		]
		lib.udev_monitor_filter_add_match_subsystem_devtype.restype = ctypes.c_int
		lib.udev_monitor_filter_add_match_tag.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
		lib.udev_monitor_filter_add_match_tag.restype = ctypes.c_int
		# device
		lib.udev_device_get_action.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_action.restype = ctypes.c_char_p
		lib.udev_device_get_devnode.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_devnode.restype = ctypes.c_char_p
		lib.udev_device_get_subsystem.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_subsystem.restype = ctypes.c_char_p
		lib.udev_device_get_devtype.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_devtype.restype = ctypes.c_char_p
		lib.udev_device_get_syspath.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_syspath.restype = ctypes.c_char_p
		lib.udev_device_get_sysname.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_sysname.restype = ctypes.c_char_p
		lib.udev_device_get_is_initialized.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_is_initialized.restype = ctypes.c_int
		lib.udev_device_get_devnum.argtypes = [ctypes.c_void_p]
		lib.udev_device_get_devnum.restype = ctypes.c_int
		lib.udev_device_unref.argtypes = [ctypes.c_void_p]

		for name in dir(Enumerator):
			if "match_" in name:
				twoargs = getattr(getattr(Enumerator, name), "twoargs", False)
				fn = getattr(lib, "udev_enumerate_add_" + name)
				if twoargs:
					fn.argtypes = [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p]
				else:
					fn.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
				fn.restype = ctypes.c_int

	def __del__(self) -> None:
		if self._ctx is not None:
			self._lib.udev_unref(self._ctx)
			self._ctx = None

	enumerate_typevar = TypeVar("enumerate_typevar", bound="Enumerator")

	def enumerate(self, subclass: type[enumerate_typevar] | None = None) -> enumerate_typevar | Enumerator:
		"""Return new Enumerator instance."""
		enumerator: ctypes.c_void_p | None = self._lib.udev_enumerate_new(self._ctx)
		if enumerator is None:
			raise OSError("Failed to initialize enumerator")
		if subclass is not None:
			assert issubclass(subclass, Enumerator), (
				f"subclass must be a subclass of Enumerator but {subclass} was provided"
			)
		# print('enumerate - ', type(enumerator), enumerator)
		final_class = subclass or Enumerator
		return final_class(self, enumerator)

	monitor_typevar = TypeVar("monitor_typevar", bound="Monitor")

	def monitor(self, subclass: type[monitor_typevar] | None = None) -> monitor_typevar | Monitor:
		"""Return new Monitor instance."""
		monitor: ctypes.c_void_p | None = self._lib.udev_monitor_new_from_netlink(self._ctx, b"udev")
		if monitor is None:
			raise OSError("Failed to initialize monitor")
		if subclass is not None:
			assert issubclass(subclass, Monitor), f"subclass must be a subclass of Monitor but {subclass} was provided"
		final_class = subclass or Monitor
		return final_class(self, monitor)


def twoargs(fn):
	fn.twoargs = True
	return fn


class Enumerator:
	"""Iterable object used for enumerating available devices.

	Yields syspaths (strings).

	All match_* methods are returning self for chaining.
	"""

	def __init__(self, eudev: Eudev, enumerator: ctypes.c_void_p) -> None:
		self._eudev = eudev
		self._enumerator = enumerator
		self._keep_in_mem = []
		self._enumeration_started = False
		self._next: ctypes.c_void_p | None = None

	def __del__(self) -> None:
		if self._enumerator is not None:
			self._eudev._lib.udev_enumerate_unref(self._enumerator)
			self._enumerator = None

	def _add_match(self, whichone: str, *pars) -> Self:
		if self._enumeration_started:
			raise RuntimeError("Cannot add match after enumeration is started")
		fn = getattr(self._eudev._lib, "udev_enumerate_add_" + whichone)
		pars = [ctypes.c_char_p(p.encode("utf-8") if type(p) is str else p) for p in pars]
		self._keep_in_mem += pars
		err = fn(self._enumerator, *pars)
		if err < 0:
			raise OSError("udev_enumerate_add_%s: error %s" % (whichone, err))
		return self

	@twoargs
	def match_sysattr(self, sysattr, value) -> Self:
		return self._add_match("match_sysattr", sysattr, value)

	@twoargs
	def nomatch_sysattr(self, sysattr, value) -> Self:
		return self._add_match("nomatch_sysattr", sysattr, value)

	@twoargs
	def match_property(self, property, value) -> Self:
		return self._add_match("match_property", property, value)

	def match_subsystem(self, subsystem: str) -> Self:
		return self._add_match("match_subsystem", subsystem)

	def nomatch_subsystem(self, subsystem: str) -> Self:
		return self._add_match("nomatch_subsystem", subsystem)

	def match_sysname(self, sysname: str) -> Self:
		return self._add_match("match_sysname", sysname)

	def match_tag(self, tag) -> Self:
		return self._add_match("match_tag", tag)

	def match_is_initialized(self) -> Self:
		return self._add_match("match_is_initialized")

	# match_parent is not implemented

	def __iter__(self) -> Self:
		if self._enumeration_started:
			raise RuntimeError("Cannot iterate same Enumerator twice")
		self._enumeration_started = True
		err = self._eudev._lib.udev_enumerate_scan_devices(self._enumerator)
		if err < 0:
			raise OSError("udev_enumerate_scan_devices: error %s" % (err,))
		self._next = self._eudev._lib.udev_enumerate_get_list_entry(self._enumerator)
		return self

	def next(self) -> str:
		return self.__next__()

	def __next__(self) -> str:
		if not self._enumeration_started:
			self.__iter__()  # Starts the enumeration
		if self._next is None:
			raise StopIteration
		udev_name_pointer: ctypes.c_char_p | None = self._eudev._lib.udev_list_entry_get_name(self._next)
		if udev_name_pointer is None:
			raise OSError("udev_list_entry_get_name failed, can't get syspath")
		self._next = self._eudev._lib.udev_list_entry_get_next(self._next)
		return str(udev_name_pointer, "utf-8")


class DeviceEvent(NamedTuple):
	action: str
	node: str | None
	initialized: bool
	subsystem: str
	devtype: str
	syspath: str
	devnum: int


class Monitor:
	"""Monitor object that receives device events.

	receive_device method blocks until next event is processed, so it can be
	used either in dumb loop, or called when select syscall reports descriptor
	returned by get_fd has data available.

	All match_* methods are returning self for chaining
	"""

	def __init__(self, eudev: Eudev, monitor: ctypes.c_void_p) -> None:
		self._eudev = eudev
		self._monitor = monitor
		self._monitor_started = False
		self._keep_in_mem = []
		self._enabled_matches = set()

	def __del__(self) -> None:
		if self._monitor is not None:
			self._eudev._lib.udev_monitor_unref(self._monitor)
			self._monitor = None

	def _add_match(self, whichone, *pars) -> Self:
		key = tuple([whichone] + list(pars))
		if key in self._enabled_matches:
			# Already done
			return self
		fn = getattr(self._eudev._lib, "udev_monitor_filter_add_" + whichone)
		pars = [ctypes.c_char_p(p.encode("utf-8") if type(p) is str else p) for p in pars]
		self._keep_in_mem += pars
		err = fn(self._monitor, *pars)
		if err < 0:
			raise OSError("udev_monitor_filter_add_%s: error %s" % (whichone, errno.errorcode.get(err, err)))
		self._enabled_matches.add(key)
		if self._monitor_started:
			err = self._eudev._lib.udev_monitor_filter_update(self._monitor)
			if err < 0:
				raise OSError("udev_monitor_filter_update: error %s" % (errno.errorcode.get(err, err),))
		return self

	def match_subsystem_devtype(self, subsystem: str, devtype=None):
		return self._add_match("match_subsystem_devtype", subsystem, devtype)

	def match_subsystem(self, subsystem: str):
		return self._add_match("match_subsystem_devtype", subsystem, None)

	def match_tag(self, tag):
		return self._add_match("match_tag", tag)

	def is_started(self):
		return self._monitor_started

	def get_fd(self):
		fileno = self._eudev._lib.udev_monitor_get_fd(self._monitor)
		if fileno < 0:
			raise OSError("udev_monitor_get_fd: error %s" % (errno.errorcode.get(fileno, fileno),))
		return fileno

	def enable_receiving(self) -> Self | None:
		"""Returns self for chaining"""
		if self._monitor_started:
			return None  # Error, but unimportant
		err = self._eudev._lib.udev_monitor_enable_receiving(self._monitor)
		if err < 0:
			raise OSError("udev_monitor_enable_receiving: error %s" % (errno.errorcode.get(err, err)))
		self._monitor_started = True
		return self

	def set_receive_buffer_size(self, size) -> Self:
		"""Returns self for chaining"""
		err = self._eudev._lib.udev_monitor_set_receive_buffer_size(self._monitor, size)
		if err < 0:
			raise OSError("udev_monitor_set_receive_buffer_size: error %s" % (errno.errorcode.get(err, err)))
		return self

	fileno = get_fd  # python stuff likes this name better
	start = enable_receiving  # I like this name better

	def receive_device(self) -> DeviceEvent | None:
		if not self._monitor_started:
			self.enable_receiving()

		dev = self._eudev._lib.udev_monitor_receive_device(self._monitor)
		if dev is None:
			# udev_monitor_receive_device is _supposed_ to be blocking.
			# It doesn't looks that way
			return None

		devnode = self._eudev._lib.udev_device_get_devnode(dev)
		devnode_str = str(devnode, "utf-8") if devnode else None

		devtype = self._eudev._lib.udev_device_get_devtype(dev)
		devtype_str = str(devtype, "utf-8") if devtype else None

		event = DeviceEvent(
			str(self._eudev._lib.udev_device_get_action(dev), "utf-8"),
			devnode_str,
			self._eudev._lib.udev_device_get_is_initialized(dev) == 1,
			str(self._eudev._lib.udev_device_get_subsystem(dev), "utf-8"),
			str(self._eudev._lib.udev_device_get_devtype(dev) or b"", "utf-8"),
			str(self._eudev._lib.udev_device_get_syspath(dev) or b"", "utf-8"),
			self._eudev._lib.udev_device_get_devnum(dev),
		)

		self._eudev._lib.udev_device_unref(dev)
		return event


if __name__ == "__main__":
	udev = Eudev()
	en = udev.enumerate().match_subsystem("hidraw")
	for i in en:
		print(i)

	m = udev.monitor().match_subsystem("hidraw").start()
	if m is None:
		msg = "Expected to receive a Monitor device, not None."
		raise RuntimeError(msg)

	while True:
		dev = m.receive_device()
		if dev is not None:
			print(f"{os.major(dev.devnum)} {os.minor(dev.devnum)} {dev}")



"""SC Controller - Profile

Handles mapping profile stored in json file
"""

from __future__ import annotations

import json
import logging
from typing import Self

from scc.actions import NoAction
from scc.constants import CPAD, DPAD, GYRO, LEFT, RIGHT, RSTICK, STICK, WHOLE, SCButtons
from scc.lib.jsonencoder import JSONEncoder
from scc.menu_data import MenuData
from scc.modifiers import HoldModifier
from scc.special_actions import MenuAction

log = logging.getLogger("profile")


class Profile:
	# Current profile version. When loading profile file with version lower than this, auto-conversion may happen
	VERSION = 1.4

	LEFT = LEFT
	RIGHT = RIGHT
	LPAD = SCButtons.LPAD.name
	RPAD = SCButtons.RPAD.name
	CPAD = CPAD
	DPAD = DPAD
	WHOLE = WHOLE
	STICK = STICK
	RSTICK = RSTICK
	GYRO = GYRO
	X, Y, Z = "X", "Y", "Z"
	STICK_AXES = {X: "lpad_x", Y: "lpad_y"}
	LPAD_AXES = STICK_AXES
	RPAD_AXES = {X: "rpad_x", Y: "rpad_y"}
	TRIGGERS = [LEFT, RIGHT]

	def __init__(self, parser) -> None:
		self.parser = parser
		self.clear()
		self.filename: str | None = None
		# UI-only values
		self.is_template: bool = False
		self.description: str = ""

	def save(self, filename: str) -> Self:
		"""Saves profile into file. Returns self"""
		with open(filename, "w") as fileobj:
			self.save_fileobj(fileobj)
		return self

	def save_fileobj(self, fileobj) -> Self:
		"""Saves profile into file-like object. Returns self"""
		data = {
			"_": (self.description if "\n" not in self.description else self.description.strip("\n").split("\n")),
			"buttons": {},
			"stick": self.stick,
			"rstick": self.rstick,
			"gyro": self.gyro,
			"trigger_left": self.triggers[Profile.LEFT],
			"trigger_right": self.triggers[Profile.RIGHT],
			"pad_left": self.pads[Profile.LEFT],
			"pad_right": self.pads[Profile.RIGHT],
			"cpad": self.pads[Profile.CPAD],
			"dpad": self.pads[Profile.DPAD],
			"menus": {id: self.menus[id].encode() for id in self.menus},
			"is_template": self.is_template,
			"version": Profile.VERSION,
		}

		for i in self.buttons:
			if self.buttons[i]:
				data["buttons"][i.name] = self.buttons[i]

		# Generate & save json
		jstr = Encoder(sort_keys=True, indent=4).encode(data)
		fileobj.write(jstr)
		return self

	def load(self, filename: str) -> Self:
		"""Loads profile from file. Returns self"""
		with open(filename) as fileobj:
			self.load_fileobj(fileobj)
		self.filename = filename
		return self

	def load_fileobj(self, fileobj) -> Self:
		"""Loads profile from file-like object.

		Filename attribute is not set, what may cause some trouble if used in GUI.

		Returns self.
		"""
		data = json.loads(fileobj.read())
		# Version
		try:
			version = float(data["version"])
		except Exception:
			logging.exception("Failed to load file-like object")
			version = 0

		# Settings - Description
		# (stored in key "_", so it's serialized on top of JSON file)
		if "_" not in data:
			self.description = ""
		elif type(data["_"]) == list:
			self.description = "\n".join(data["_"])
		else:
			self.description = data["_"]
		# Settings - Template
		self.is_template = bool(data["is_template"]) if "is_template" in data else False

		# Buttons
		self.buttons = {}
		for x in SCButtons:
			self.buttons[x] = self.parser.from_json_data(data["buttons"], x.name)
		# Pressing stick is interpreted as STICKPRESS button,
		# formely called just STICK
		if "STICK" in data["buttons"] and "STICKPRESS" not in data["buttons"]:
			self.buttons[SCButtons.STICKPRESS] = self.parser.from_json_data(data["buttons"], "STICK")

		# Stick & gyro
		self.stick = self.parser.from_json_data(data, "stick")
		self.gyro = self.parser.from_json_data(data, "gyro")

		if "triggers" in data:
			# Old format
			# Triggers
			self.triggers = {x: self.parser.from_json_data(data["triggers"], x) for x in Profile.TRIGGERS}

			# Pads
			self.pads = {
				Profile.LEFT: self.parser.from_json_data(data, "left_pad"),
				Profile.RIGHT: self.parser.from_json_data(data, "right_pad"),
				Profile.CPAD: NoAction(),
				Profile.DPAD: NoAction(),
			}

			# Rigth stick
			self.rstick = NoAction()
		else:
			# New format
			# Triggers
			self.triggers = {
				Profile.LEFT: self.parser.from_json_data(data, "trigger_left"),
				Profile.RIGHT: self.parser.from_json_data(data, "trigger_right"),
			}

			# Pads
			self.pads = {
				Profile.LEFT: self.parser.from_json_data(data, "pad_left"),
				Profile.RIGHT: self.parser.from_json_data(data, "pad_right"),
				Profile.CPAD: self.parser.from_json_data(data, "cpad"),
				Profile.DPAD: self.parser.from_json_data(data, "dpad"),
			}

			# Rigth stick
			self.rstick = self.parser.from_json_data(data, "rstick")

		# Menus
		self.menus = {}
		if "menus" in data:
			for id in data["menus"]:
				for invalid_char in ".:/":
					if invalid_char in id:
						raise ValueError("Invalid character '%s' in menu id '%s'" % (invalid_char, id))
				self.menus[id] = MenuData.from_json_data(data["menus"][id], self.parser)

		# Conversion
		self.original_version = version  # TODO: This is temporary
		if version < Profile.VERSION:
			self._convert(version)

		return self

	def clear(self) -> None:
		"""Clears all actions and adds default menu action on center button"""
		self.buttons = {x: NoAction() for x in SCButtons}
		self.buttons[SCButtons.C] = HoldModifier(MenuAction("Default.menu"), normalaction=MenuAction("Default.menu"))
		self.menus = {}
		self.stick = NoAction()
		self.rstick = NoAction()
		self.is_template = False
		self.triggers = {Profile.LEFT: NoAction(), Profile.RIGHT: NoAction()}
		self.pads = {
			Profile.LEFT: NoAction(),
			Profile.RIGHT: NoAction(),
			Profile.CPAD: NoAction(),
			Profile.DPAD: NoAction(),
		}
		self.gyro = NoAction()

	def get_all_actions(self):
		"""Returns generator with every action defined in this profile,

		including actions in menus.
		Recursively walks into macros, dpads and everything else that can have
		nested actions, so both parent and all child actions are yielded.

		May yield NoAction, but shouldn't yield None.

		Used for checks when profile is exported or imported.
		"""
		for action in self.get_actions():
			for i in action.get_all_actions():
				yield i
		for id in self.menus:
			for i in self.menus[id].get_all_actions():
				yield i

	def get_actions(self):
		"""As get_all_actions, but returns only root actions, without children, and ignores menus."""
		for dct in (self.buttons, self.triggers, self.pads):
			for k in dct:
				yield dct[k]
		for action in (self.stick, self.rstick, self.gyro):
			yield action

	def get_filename(self):
		"""Returns filename of last loaded file or None."""
		return self.filename

	def compress(self) -> None:
		"""Calls compress on every action to throw out some redundant stuff.

		Note that calling save() after compress() will cause data loss.
		"""
		for dct in (self.buttons, self.triggers, self.pads):
			for x in dct:
				dct[x] = dct[x].compress()
		self.rstick = self.rstick.compress()
		self.stick = self.stick.compress()
		self.gyro = self.gyro.compress()
		for menu in self.menus.values():
			menu.compress()

	def _convert(self, from_version):
		"""Performs conversion from older profile version"""
		if from_version < 1:

			# Add 'display Default.menu if center button is held' for old profiles
			c = self.buttons[SCButtons.C]
			if not c:
				# Nothing set to C button
				self.buttons[SCButtons.C] = HoldModifier(
					MenuAction("Default.menu"), normalaction=MenuAction("Default.menu"),
				)
			elif hasattr(c, "holdaction") and c.holdaction:
				# Already set to something, don't overwrite it
				pass
			elif c.to_string().startswith("OSK."):
				# Special case, don't touch this either
				pass
			else:
				self.buttons[SCButtons.C] = HoldModifier(
					MenuAction("Default.menu"), normalaction=self.buttons[SCButtons.C],
				)
		if from_version < 1.1:
			# Convert old scrolling wheel to new representation
			from scc.actions import MouseAction, XYAction
			from scc.modifiers import BallModifier, FeedbackModifier
			from scc.uinput import Rels

			iswheelaction = lambda x: isinstance(x, MouseAction) and x.parameters[0] in (
				Rels.REL_HWHEEL,
				Rels.REL_WHEEL,
			)
			for p in (Profile.LEFT, Profile.RIGHT):
				a, feedback = self.pads[p], None
				if isinstance(a, FeedbackModifier):
					feedback = a.haptic.get_position()
					a = a.action
				if isinstance(a, XYAction):
					if iswheelaction(a.x) or iswheelaction(a.y):
						n = BallModifier(XYAction(a.x, a.y))
						if feedback is not None:
							n = FeedbackModifier(feedback, 4096, 16, n)
						self.pads[p] = n
						log.info("Converted %s to %s", a.to_string(), n.to_string())
		if from_version < 1.2:
			# Convert old trigger settings that were done with ButtonAction
			# to new TriggerAction
			from scc.actions import ButtonAction, MultiAction, TriggerAction
			from scc.constants import TRIGGER_CLICK, TRIGGER_HALF, TRIGGER_MAX
			from scc.uinput import Keys

			for p in (Profile.LEFT, Profile.RIGHT):
				if isinstance(self.triggers[p], ButtonAction):
					buttons, numbers = [], []
					n = None
					# There were one or two keys and zero to two numeric
					# parameters for old button action
					for param in self.triggers[p].parameters:
						if param in Keys:
							buttons.append(param)
						elif type(param) in (int, float):
							numbers.append(int(param))
					if len(numbers) == 0:
						# Trigger range was not specified, assume defaults
						numbers = (TRIGGER_HALF, TRIGGER_CLICK)
					elif len(numbers) == 1:
						# Only lower range was specified, add default upper range
						numbers.append(TRIGGER_CLICK)
					if len(buttons) == 1:
						# If only one button was set, trigger should work like
						# one big button
						n = TriggerAction(numbers[0], ButtonAction(buttons[0]))
					elif len(buttons) == 2:
						# Both buttons were set
						n = MultiAction(
							TriggerAction(numbers[0], numbers[1], ButtonAction(buttons[0])),
							TriggerAction(numbers[1], TRIGGER_MAX, ButtonAction(buttons[1])),
						)

					if n:
						log.info("Converted %s to %s", self.triggers[p].to_string(), n.to_string())
						self.triggers[p] = n
		if from_version < 1.3:
			# Action format completly changed in v0.4, but profile format is same.
			pass


class Encoder(JSONEncoder):
	def default(self, obj):
		# if type(obj) in (list, tuple):
		# return basestring("[" + ", ".join(self.encode(x) for x in obj) + " ]")
		if hasattr(obj, "encode"):
			return obj.encode()
		return JSONEncoder.default(self, obj)



#!/usr/bin/env python3
"""SC Controller - Macros

Frontier is my favorite.
"""

import logging

from scc.actions import Action, ButtonAction
from scc.uinput import Keys

log = logging.getLogger("Macros")
_ = lambda x: x


class Macro(Action):
	"""Two or more actions executed in sequence.
	Generated when parsing ';'
	"""

	COMMAND = None
	HOLD_TIME = 0.01

	def __init__(self, *parameters):
		Action.__init__(self, *parameters)
		self.actions = []
		self.repeat = False
		self.hold_time = Macro.HOLD_TIME
		self._active = False
		self._current = None
		self._release = None
		for p in parameters:
			if type(p) == float and len(self.actions):
				self.actions[-1].delay_after = p
			elif isinstance(p, Macro):
				self.actions += p.actions
			elif isinstance(p, Action):
				self.actions.append(p)
			else:
				self.actions.append(ButtonAction(p))

	def button_press(self, mapper):
		# Macro can be executed only by pressing button
		if len(self.actions) < 1:
			# Empty macro
			return False
		self._active = True
		if self._current is not None:
			# Already executing macro
			return False
		self._current = [] + self.actions
		self.timer(mapper)

	def timer(self, mapper):
		if self._release is None:
			# Execute next action
			self._release, self._current = self._current[0], self._current[1:]
			self._release.button_press(mapper)
			mapper.schedule(self.hold_time, self.timer)
		else:
			# Finish execited action
			self._release.button_release(mapper)
			if len(self._current) == 0 and self.repeat and self._active:
				# Repeating
				self._current = [] + self.actions
				mapper.schedule(self._release.delay_after, self.timer)
				self._release = None
			elif len(self._current) == 0:
				# Finished
				self._current = None
				self._release = None
			else:
				# Schedule for next action
				mapper.schedule(self._release.delay_after, self.timer)
				self._release = None

	def cancel(self, mapper):
		for a in self.actions:
			a.cancel(mapper)

	def set_haptic(self, hapticdata):
		for a in self.actions:
			if a and hasattr(a, "set_haptic"):
				a.set_haptic(hapticdata)

	def get_haptic(self):
		for a in self.actions:
			if a and hasattr(a, "set_haptic"):
				return a.get_haptic()
		return None

	def set_speed(self, x, y, z):
		for a in self.actions:
			if hasattr(a, "set_speed"):
				a.set_speed(x, y, z)

	def get_speed(self):
		for a in self.actions:
			if hasattr(a, "set_speed"):
				return a.get_speed()
		return (1.0,)

	def button_release(self, mapper):
		self._active = False

	def describe(self, context):
		if self.name:
			return self.name
		if self.repeat:
			return "repeat " + "; ".join([x.describe(context) for x in self.actions])
		return "; ".join([x.describe(context) for x in self.actions])

	def to_string(self, multiline=False, pad=0):
		lst = "; ".join([x.to_string() for x in self.actions])
		if self.repeat:
			return (" " * pad) + ("repeat(%s)" % (lst,))
		return (" " * pad) + lst

	def __str__(self):
		if self.repeat:
			return "<[repeat %s ]>" % ("; ".join([str(x) for x in self.actions]),)
		return "<[ %s ]>" % ("; ".join([str(x) for x in self.actions]),)

	__repr__ = __str__


class Type(Macro):
	"""Special type of Macro where keys to press are specified as string.
	Basically, writing type("iddqd") is same thing as
	button(KEY_I) ; button(KEY_D) ; button(KEY_D); button(KEY_Q); button(KEY_D)

	Recognizes only lowercase letters, uppercase letters, numbers and space.
	Adding anything else will make action unparseable.
	"""

	COMMAND = "type"
	HOLD_TIME = 0.001

	def __init__(self, string):
		params = []
		shift = False
		for letter in string:
			if (letter >= "a" and letter <= "z") or (letter >= "0" and letter <= "9"):
				if hasattr(Keys, ("KEY_" + letter).upper()):
					if shift:
						params.append(ReleaseAction(Keys.KEY_LEFTSHIFT))
						shift = False
					params.append(ButtonAction(getattr(Keys, ("KEY_" + letter).upper())))
					continue
			if letter == " ":
				params.append(ButtonAction(Keys.KEY_SPACE))
				continue
			if letter >= "A" and letter <= "Z":
				if hasattr(Keys, "KEY_" + letter):
					if not shift:
						params.append(PressAction(Keys.KEY_LEFTSHIFT))
						shift = True
					params.append(ButtonAction(getattr(Keys, "KEY_" + letter)))
					continue
			raise ValueError("Invalid character for type(): '%s'" % (letter,))
		Macro.__init__(self, *params)
		self.letters = string

	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + self.COMMAND + "(" + repr(self.letters).strip("u") + ")"


class Cycle(Macro):
	"""Multiple actions cycling on same button.
	When button is pressed 1st time, 1st action is executed. 2nd action is
	executed for 2nd press et cetera et cetera.
	"""

	COMMAND = "cycle"

	def __init__(self, *parameters):
		Action.__init__(self, *parameters)
		self.actions = parameters
		self._current = 0

	def button_press(self, mapper):
		if len(self.actions) > 0:
			self.actions[self._current].button_press(mapper)

	def button_release(self, mapper):
		if len(self.actions) > 0:
			self.actions[self._current].button_release(mapper)
			self._current += 1
			if self._current >= len(self.actions):
				self._current = 0

	def describe(self, context):
		if self.name:
			return self.name
		return _("Cycle Actions")

	def to_string(self, multiline=False, pad=0):
		lst = ", ".join([x.to_string() for x in self.actions])
		return (" " * pad) + self.COMMAND + "(" + lst + ")"

	def __str__(self):
		return "<cycle %s >" % ("; ".join([str(x) for x in self.actions]),)

	__repr__ = __str__


class Repeat(Macro):
	"""Repeats specified action as long as physical button is pressed.
	This is actually just Macro with 'repeat' set to True
	"""

	COMMAND = "repeat"

	def __new__(cls, action):
		if not isinstance(action, Macro):
			action = Macro(action)
		action.repeat = True
		return action


class SleepAction(Action):
	"""Does nothing.
	If used in macro, overrides delay after itself.
	"""

	COMMAND = "sleep"

	def __init__(self, delay):
		Action.__init__(self, delay)
		self.delay = float(delay)
		self.delay_after = self.delay - Macro.HOLD_TIME

	def describe(self, context):
		if self.name:
			return self.name
		if self.delay < 1.0:
			return _("Wait %sms") % (int(self.delay * 1000),)
		s = ("%0.2f" % (self.delay,)).strip(".0")
		return _("Wait %ss") % (s,)

	def to_string(self, multiline=False, pad=0):
		return (" " * pad) + "%s(%0.3f)" % (self.COMMAND, self.delay)

	def button_press(self, mapper):
		pass

	def button_release(self, mapper):
		pass


class PressAction(Action):
	"""Presses button and leaves it pressed.
	Can be used anywhere, but makes sense only with macro.
	"""

	COMMAND = "press"
	PR = _("Press")

	def __init__(self, action):
		Action.__init__(self, action)
		self.action = action

	def describe_short(self):
		"""Used in macro editor"""
		if isinstance(self.action, ButtonAction):
			return self.action.describe_short()
		if isinstance(self.action, Keys):
			return ButtonAction.describe_button(self.action)
		return self.action.describe(Action.AC_BUTTON)

	def describe(self, context):
		if self.name:
			return self.name
		return self.PR + " " + self.describe_short()

	def button_press(self, mapper):
		self.action.button_press(mapper)

	def button_release(self, mapper):
		# This is activated only when button is pressed
		pass


class ReleaseAction(PressAction):
	"""Releases button.
	Can be used anywhere, but makes sense only with macro.
	"""

	COMMAND = "release"
	PR = _("Release")

	def button_press(self, mapper):
		self.action.button_release(mapper)


class TapAction(PressAction):
	"""Presses button for short time.
	If button is already pressed, generates release-press-release-press
	events in quick sequence.
	"""

	COMMAND = "tap"
	PR = _("Tap")
	PAUSE = 0.1
	COUNTER_VAL = 100

	def __init__(self, button, count=1):
		PressAction.__init__(self, button)
		self._lst = []
		self._keep_pressed = False
		self.button = button
		self.count = count

	def button_press(self, mapper):
		if len(self._lst):
			# Still executing from scheduler
			return

		# ---
		# This thing abuses internal "button press" counter a little; First,
		# if button is supposedly pressed more than 1 times (because two or more
		# actions are holding it down at same time), tap is aborted. That is not
		# ideal, but shouldn't be a problem in most cases.
		#
		# 2nd, counter is manipulated before every call to _button_press / release,
		# so it always fires emulated button. Then, while scheduler is active,
		# counter is bumped to 'COUNTER_VAL', so it can be detected that other
		# action touched same button.
		# ---
		if self.button in mapper.pressed and mapper.pressed[self.button] > 1:
			log.warning("Failed to tap, two or more actions are holding button")
			return

		# Generate as many clicks as requested
		# True is for press, False for release
		self._lst = [True, False] * self.count

		if self.button in mapper.pressed and mapper.pressed[self.button] > 0:
			# Surround by release - ... - press if button is currently pressed
			self._lst = [False] + self._lst + [True]
		elif self.count > 1:
			# Keep button pressed if double-or-more tap was requested
			self._lst = self._lst[0:-1]
			self._keep_pressed = True

		mapper.pressed[self.button] = self.COUNTER_VAL
		self._rel_tap_press(mapper)

	def _bailout(self):
		self._lst, self._keep_pressed = [], None

	def _rel_tap_press(self, mapper):
		if self.button not in mapper.pressed or mapper.pressed[self.button] < self.COUNTER_VAL:
			# Something else tried to _release_ button in meanwhile, bail out
			mapper.pressed[self.button] = 1
			ButtonAction._button_release(mapper, self.button)
			return self._bailout()
		if mapper.pressed[self.button] > self.COUNTER_VAL:
			# Something else pressed button in meanwhile, bail out
			mapper.pressed[self.button] = 1
			return self._bailout()

		a, self._lst = self._lst[0], self._lst[1:]
		if a:
			mapper.pressed[self.button] = 0
			ButtonAction._button_press(mapper, self.button)
		else:
			mapper.pressed[self.button] = 1
			ButtonAction._button_release(mapper, self.button)
		if len(self._lst):
			mapper.pressed[self.button] = self.COUNTER_VAL
			mapper.schedule(self.PAUSE, self._rel_tap_press)

	def button_release(self, mapper):
		if self._keep_pressed:
			self._keep_pressed = False
			if len(self._lst) > 0:
				# _rel_tap_press is still scheduled
				self._lst += [False]
			else:
				ButtonAction._button_release(mapper, self.button)

	def describe_short(self):
		"""Used in macro editor"""
		if self.count <= 1:
			return "%s %s" % (_("Tap"), ButtonAction.describe_button(self.button))
		if self.count == 2:
			return "%s %s" % (_("DblTap"), ButtonAction.describe_button(self.button))
		return "%s%s %s" % (self.count, _("-tap"), ButtonAction.describe_button(self.button))

	def describe(self, context):
		if self.name:
			return self.name
		return self.describe_short()

	def to_string(self, multiline=False, pad=0):
		if self.count <= 1:
			return "%s(%s)" % (self.COMMAND, self.button)
		return "%s(%s, %s)" % (self.COMMAND, self.button, self.count)



#!/usr/bin/env python3
"""SC-Controller - AREA_TO_ACTION

Maps areas on SVG images into actions.
Used by ActionEditor.
"""

from scc.actions import (
	AxisAction,
	ButtonAction,
	HatDownAction,
	HatLeftAction,
	HatRightAction,
	HatUpAction,
	MouseAction,
	RAxisAction,
)
from scc.uinput import Axes, Keys, Rels

AREA_TO_ACTION = {
	# Values in tuples: ActionClass, param1, param2...
	# Buttons
	"TL": (ButtonAction, Keys.BTN_TL),
	"TR": (ButtonAction, Keys.BTN_TR),
	"THUMBL": (ButtonAction, Keys.BTN_THUMBL),
	"THUMBR": (ButtonAction, Keys.BTN_THUMBR),
	"SELECT": (ButtonAction, Keys.BTN_SELECT),
	"MODE": (ButtonAction, Keys.BTN_MODE),
	"START": (ButtonAction, Keys.BTN_START),
	"A": (ButtonAction, Keys.BTN_A),
	"B": (ButtonAction, Keys.BTN_B),
	"X": (ButtonAction, Keys.BTN_X),
	"Y": (ButtonAction, Keys.BTN_Y),
	# Media keys
	"KEY_PREVIOUSSONG": (ButtonAction, Keys.KEY_PREVIOUSSONG),
	"KEY_STOP": (ButtonAction, Keys.KEY_STOP),
	"KEY_PLAYPAUSE": (ButtonAction, Keys.KEY_PLAYPAUSE),
	"KEY_NEXTSONG": (ButtonAction, Keys.KEY_NEXTSONG),
	"KEY_VOLUMEDOWN": (ButtonAction, Keys.KEY_VOLUMEDOWN),
	"KEY_VOLUMEUP": (ButtonAction, Keys.KEY_VOLUMEUP),
	# Dpad
	"DPAD_LEFT": (HatLeftAction, Axes.ABS_HAT0X),
	"DPAD_RIGHT": (HatRightAction, Axes.ABS_HAT0X),
	"DPAD_UP": (HatUpAction, Axes.ABS_HAT0Y),
	"DPAD_DOWN": (HatDownAction, Axes.ABS_HAT0Y),
	"ABS_HAT0X": (AxisAction, Axes.ABS_HAT0X),
	"ABS_HAT0Y": (AxisAction, Axes.ABS_HAT0Y),
	# Left stick
	"LSTICK_LEFT": (AxisAction, Axes.ABS_X, 0, -32767),
	"LSTICK_RIGHT": (AxisAction, Axes.ABS_X, 0, 32767),
	"LSTICK_UP": (AxisAction, Axes.ABS_Y, 0, -32767),
	"LSTICK_DOWN": (AxisAction, Axes.ABS_Y, 0, 32767),
	"ABS_X": (AxisAction, Axes.ABS_X),
	"ABS_Y": (AxisAction, Axes.ABS_Y),
	# Right stick
	"RSTICK_LEFT": (AxisAction, Axes.ABS_RX, 0, -32767),
	"RSTICK_RIGHT": (AxisAction, Axes.ABS_RX, 0, 32767),
	"RSTICK_UP": (AxisAction, Axes.ABS_RY, 0, 32767),
	"RSTICK_DOWN": (AxisAction, Axes.ABS_RY, 0, -32767),
	"ABS_RX": (AxisAction, Axes.ABS_RX),
	"ABS_RY": (AxisAction, Axes.ABS_RY),
	# Halves
	"MINUSHALF_X": (AxisAction, Axes.ABS_X, 0, -32767),
	"PLUSHALF_X": (AxisAction, Axes.ABS_X, 0, 32767),
	"MINUSHALF_Y": (AxisAction, Axes.ABS_Y, 0, -32767),
	"PLUSHALF_Y": (AxisAction, Axes.ABS_Y, 0, 32767),
	"MINUSHALF_RX": (AxisAction, Axes.ABS_RX, 0, -32767),
	"PLUSHALF_RX": (AxisAction, Axes.ABS_RX, 0, 32767),
	"MINUSHALF_RY": (AxisAction, Axes.ABS_RY, 0, -32767),
	"PLUSHALF_RY": (AxisAction, Axes.ABS_RY, 0, 32767),
	# Triggers
	"ABS_Z": (AxisAction, Axes.ABS_Z),
	"ABS_RZ": (AxisAction, Axes.ABS_RZ),
	# Mouse
	"MOUSE_LEFT": (MouseAction, Rels.REL_X, -1),
	"MOUSE_RIGHT": (MouseAction, Rels.REL_X, 1),
	"MOUSE_UP": (MouseAction, Rels.REL_Y, -1),
	"MOUSE_DOWN": (
		MouseAction,
		Rels.REL_Y,
		1,
	),
	"MOUSE_X": (MouseAction, Rels.REL_X, 1),
	"MOUSE_Y": (MouseAction, Rels.REL_Y, 1),
	"MOUSE_WHEEL": (MouseAction, Rels.REL_WHEEL, 1),
	"MOUSE_HWHEEL": (MouseAction, Rels.REL_HWHEEL, 1),
	# Mouse buttons
	"MOUSE1": (ButtonAction, Keys.BTN_LEFT),
	"MOUSE2": (ButtonAction, Keys.BTN_MIDDLE),
	"MOUSE3": (ButtonAction, Keys.BTN_RIGHT),
	"MOUSE4": (MouseAction, Rels.REL_WHEEL, 1),
	"MOUSE5": (MouseAction, Rels.REL_WHEEL, -1),
	"MOUSE8": (ButtonAction, Keys.BTN_SIDE),
	"MOUSE9": (ButtonAction, Keys.BTN_EXTRA),
}

_CLS_TO_AREA = {}
for x in AREA_TO_ACTION:
	cls, params = AREA_TO_ACTION[x][0], AREA_TO_ACTION[x][1:]
	if cls not in _CLS_TO_AREA:
		_CLS_TO_AREA[cls] = []
	_CLS_TO_AREA[cls].append((x, params))


def action_to_area(action):
	"""Returns area that matches provided action (both class and parameters)
	or None if there is no such area.
	"""
	cls = action.__class__
	if cls == RAxisAction:
		cls = AxisAction
	if cls not in _CLS_TO_AREA:
		return None
	for area, pars in _CLS_TO_AREA[cls]:
		if len(pars) > len(action.parameters):
			continue
		differs = False
		for i in range(len(pars)):
			if pars[i] != action.parameters[i]:
				differs = True
				break
		if differs:
			continue
		return area
	return None



"""SC Controller - Steam Controller Driver.

Called and used when single Steam Controller is connected directly by USB cable.

Shares a lot of classes with sc_dongle.py
"""

import logging
import struct

from usb1 import USBError

from scc.drivers.usb import USBDevice, register_hotplug_device

from .sc_dongle import TUP_FORMAT, ControllerInput, SCController, SCStatus

VENDOR_ID = 0x28DE
PRODUCT_ID = 0x1102
ENDPOINT = 3
CONTROLIDX = 2
TIMER_INTERVAL = 0.01

log = logging.getLogger("SCCable")


def init(daemon, config):
	"""Register hotplug callback for controller dongle."""

	def cb(device, handle):
		return SCByCable(device, handle, daemon)

	register_hotplug_device(cb, VENDOR_ID, PRODUCT_ID)
	return True


class SCByCable(USBDevice, SCController):
	FORMAT1 = b">BBBBB13sB2s"

	def __init__(self, device, handle, daemon):
		self.daemon = daemon
		USBDevice.__init__(self, device, handle)
		SCController.__init__(self, self, CONTROLIDX, ENDPOINT)
		self._ready = False
		self._last_tup = None
		daemon.add_mainloop(self._timer)

		self.claim_by(klass=3, subclass=0, protocol=0)
		self.read_serial()

	def generate_serial(self):
		self._serial = "%s:%s" % (self.device.getBusNumber(), self.device.getPortNumber())

	def disconnected(self):
		# Overrided to skip returning serial# to pool.
		pass

	def __repr__(self):
		return "<SCByCable %s>" % (self.get_id(),)

	def on_serial_got(self):
		log.debug("Got wired SC with serial %s", self._serial)
		self._id = "sc%s" % (self._serial,)
		self.set_input_interrupt(ENDPOINT, 64, self._wait_input)

	def _wait_input(self, endpoint, data):
		tup = ControllerInput._make(struct.unpack(TUP_FORMAT, data))
		if not self._ready:
			self.daemon.add_controller(self)
			self.configure()
			self._ready = True
		if tup.status == SCStatus.INPUT:
			self._last_tup = tup

	def _timer(self):
		m = self.get_mapper()
		if m:
			if self._last_tup:
				self.input(self._last_tup)
				self._last_tup = None
			else:
				m.generate_events()
				m.generate_feedback()
			try:
				self.flush()
			except USBError as e:
				log.exception(e)
				log.error("Error while communicating with device, baling out...")
				self.force_restart()

	def close(self):
		if self._ready:
			self.daemon.remove_controller(self)
			self._ready = False
		self.daemon.remove_mainloop(self._timer)
		USBDevice.close(self)

	def turnoff(self):
		log.warning("Ignoring request to turn off wired controller.")
